# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Miscellaneous package internals. '''



from . import imports as __


def install_builtin_safely(
    alias: str, installee: __.typx.Any, errorclass: type[ Exception ]
) -> None:
    ''' Adds attribute to interpreter builtins, if absent. '''
    import builtins
    if hasattr( builtins, alias ): raise errorclass( builtins, alias )
    setattr( builtins, alias, installee )
